import os
from collections.abc import AsyncIterator, Callable, Iterable, Iterator, Sequence
from typing import Any, Literal, cast

import openai
from openai.types.chat import (
    ChatCompletionChunk,
    ChatCompletionNamedToolChoiceParam,
    ChatCompletionStreamOptionsParam,
)

from magentic._parsing import contains_parallel_function_call_type, contains_string_type
from magentic.chat_model.base import ChatModel, OutputT, aparse_stream, parse_stream
from magentic.chat_model.function_schema import (
    get_async_function_schemas,
    get_function_schemas,
)
from magentic.chat_model.message import AssistantMessage, ContentT, Message, Usage
from magentic.chat_model.openai_chat_model import (
    BaseFunctionToolSchema,
    OpenaiChatModel,
    OpenaiStreamParser,
    OpenaiStreamState,
    _add_missing_tool_calls_responses,
    _if_given,
    async_message_to_openai_message,
    message_to_openai_message,
)
from magentic.chat_model.stream import AsyncOutputStream, OutputStream


class OpenRouterStreamState(OpenaiStreamState):
    """State for OpenRouter stream parsing."""

    reasoning: str

    def __init__(self) -> None:
        super().__init__()
        self.reasoning = ""

    def update(self, item: ChatCompletionChunk) -> None:
        super().update(item)
        if (
            hasattr(item.choices[0].delta, "reasoning")
            and item.choices[0].delta.reasoning
        ):
            self.reasoning += item.choices[0].delta.reasoning


class OpenRouterAssistantMessage(AssistantMessage[ContentT]):
    """An assistant message from OpenRouter that includes reasoning tokens."""

    reasoning: str = ""

    def __init__(self, content: ContentT, reasoning: str = "", **data: Any):
        super().__init__(content=content, **data)
        self.reasoning = reasoning

    @classmethod
    def _with_usage(
        cls,
        content: ContentT,  # type: ignore[misc]
        usage_ref: list[Usage],
        reasoning: str = "",
    ) -> "OpenRouterAssistantMessage[ContentT]":
        """Create a message with usage statistics."""
        message = cls(content=content, reasoning=reasoning)
        message._usage_ref = usage_ref
        return message


class _OpenRouterOpenaiChatModel(OpenaiChatModel):
    """Modified OpenaiChatModel to be compatible with OpenRouter API."""

    def __init__(
        self,
        model: str,
        *,
        api_key: str | None = None,
        base_url: str | None = "https://openrouter.ai/api/v1",
        max_tokens: int | None = None,
        seed: int | None = None,
        temperature: float | None = None,
        # Routing options
        route: Literal["fallback"] | None = None,
        models: list[str] | None = None,
        # Reasoner model options
        reasoning_effort: Literal["low", "medium", "high"] | None = None,
        reasoning_exclude: bool | None = None,
        # Provider options
        require_parameters: bool | None = None,
        provider_order: list[str] | None = None,
        allow_fallbacks: bool | None = None,
        data_collection: Literal["allow", "deny"] | None = None,
        provider_only: list[str] | None = None,
        provider_ignore: list[str] | None = None,
        quantizations: list[str] | None = None,
        provider_sort: Literal["price", "throughput", "latency"] | None = None,
        max_price: dict[str, float] | None = None,
    ):
        super().__init__(
            model,
            api_key=api_key,
            base_url=base_url,
            max_tokens=max_tokens,
            seed=seed,
            temperature=temperature,
        )
        self._route = route
        self._models = models
        self._reasoning_effort = reasoning_effort
        self._reasoning_exclude = reasoning_exclude
        # Provider options
        self._require_parameters = require_parameters
        self._provider_order = provider_order
        self._allow_fallbacks = allow_fallbacks
        self._data_collection = data_collection
        self._provider_only = provider_only
        self._provider_ignore = provider_ignore
        self._quantizations = quantizations
        self._provider_sort = provider_sort
        self._max_price = max_price

    def _get_stream_options(self) -> ChatCompletionStreamOptionsParam | openai.NotGiven:
        return {"include_usage": True}

    @staticmethod
    def _get_tool_choice(
        *,
        tool_schemas: Sequence[BaseFunctionToolSchema[Any]],
        output_types: Iterable[type],
    ) -> (
        Literal["none", "auto", "required"]
        | openai.NotGiven
        | ChatCompletionNamedToolChoiceParam
    ):
        if contains_string_type(output_types):
            return openai.NOT_GIVEN
        if len(tool_schemas) == 1:
            return tool_schemas[0].as_tool_choice()
        return "required"

    def _get_parallel_tool_calls(
        self, *, tools_specified: bool, output_types: Iterable[type]
    ) -> bool | openai.NotGiven:
        if not tools_specified:
            return openai.NOT_GIVEN
        if contains_parallel_function_call_type(output_types):
            return openai.NOT_GIVEN
        return False

    def _get_extra_body(self) -> dict[str, Any] | None:
        """Get extra body parameters for OpenRouter API."""
        extra_body: dict[str, Any] = {}
        if self._route:
            extra_body["route"] = self._route
        if self._models:
            extra_body["models"] = self._models

        # Build provider object
        provider: dict[str, Any] = {}
        if self._require_parameters:
            provider["require_parameters"] = True
        if self._provider_order:
            provider["order"] = self._provider_order
        if self._allow_fallbacks:
            provider["allow_fallbacks"] = True
        if self._data_collection:
            provider["data_collection"] = self._data_collection
        if self._provider_only:
            provider["only"] = self._provider_only
        if self._provider_ignore:
            provider["ignore"] = self._provider_ignore
        if self._quantizations:
            provider["quantizations"] = self._quantizations
        if self._provider_sort:
            provider["sort"] = self._provider_sort
        if self._max_price:
            provider["max_price"] = self._max_price

        if provider:
            extra_body["provider"] = provider

        # Build reasoning object
        reasoning: dict[str, Any] = {}
        if self._reasoning_effort:
            reasoning["effort"] = self._reasoning_effort
        if self._reasoning_exclude is not None:
            reasoning["exclude"] = self._reasoning_exclude
        if reasoning:
            extra_body["reasoning"] = reasoning

        return extra_body if extra_body else None

    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[OutputT]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> OpenRouterAssistantMessage[OutputT]:
        """Request an LLM message."""
        if output_types is None:
            output_types = cast("Iterable[type[OutputT]]", [] if functions else [str])

        function_schemas = get_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        response: Iterator[ChatCompletionChunk] = self._client.chat.completions.create(
            model=self.model,
            messages=_add_missing_tool_calls_responses(
                [message_to_openai_message(m) for m in messages]
            ),
            max_tokens=_if_given(self.max_tokens),
            seed=_if_given(self.seed),
            stop=_if_given(stop),
            stream=True,
            stream_options=self._get_stream_options(),
            temperature=_if_given(self.temperature),
            tools=[schema.to_dict() for schema in tool_schemas] or openai.NOT_GIVEN,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),
            parallel_tool_calls=self._get_parallel_tool_calls(
                tools_specified=bool(tool_schemas), output_types=output_types
            ),
            extra_body=self._get_extra_body(),
        )
        stream = OutputStream(
            response,
            function_schemas=function_schemas,
            parser=OpenaiStreamParser(),
            state=OpenRouterStreamState(),
        )
        return OpenRouterAssistantMessage._with_usage(
            parse_stream(stream, output_types),
            usage_ref=stream.usage_ref,
            reasoning=stream._state.reasoning if stream._state.reasoning else "",  # type: ignore[attr-defined]
        )

    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[OutputT]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> OpenRouterAssistantMessage[OutputT]:
        """Async version of `complete`."""
        if output_types is None:
            output_types = [] if functions else cast("list[type[OutputT]]", [str])

        function_schemas = get_async_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        response: AsyncIterator[
            ChatCompletionChunk
        ] = await self._async_client.chat.completions.create(
            model=self.model,
            messages=_add_missing_tool_calls_responses(
                [await async_message_to_openai_message(m) for m in messages]
            ),
            max_tokens=_if_given(self.max_tokens),
            seed=_if_given(self.seed),
            stop=_if_given(stop),
            stream=True,
            stream_options=self._get_stream_options(),
            temperature=_if_given(self.temperature),
            tools=[schema.to_dict() for schema in tool_schemas] or openai.NOT_GIVEN,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),
            parallel_tool_calls=self._get_parallel_tool_calls(
                tools_specified=bool(tool_schemas), output_types=output_types
            ),
            extra_body=self._get_extra_body(),
        )
        stream = AsyncOutputStream(
            response,
            function_schemas=function_schemas,
            parser=OpenaiStreamParser(),
            state=OpenRouterStreamState(),
        )
        return OpenRouterAssistantMessage._with_usage(
            await aparse_stream(stream, output_types),
            usage_ref=stream.usage_ref,
            reasoning=stream._state.reasoning if stream._state.reasoning else "",  # type: ignore[attr-defined]
        )


class OpenRouterChatModel(ChatModel):
    """An LLM chat model that uses OpenRouter's API via the `openai` python package."""

    def __init__(
        self,
        model: str,
        *,
        api_key: str | None = None,
        base_url: str | None = "https://openrouter.ai/api/v1",
        max_tokens: int | None = None,
        seed: int | None = None,
        temperature: float | None = None,
        # Routing options
        route: Literal["fallback"] | None = None,
        models: list[str] | None = None,
        # Reasoner model options
        reasoning_effort: Literal["low", "medium", "high"] | None = None,
        reasoning_exclude: bool | None = None,
        # Provider options
        require_parameters: bool | None = None,
        provider_order: list[str] | None = None,
        allow_fallbacks: bool | None = None,
        data_collection: Literal["allow", "deny"] | None = None,
        provider_only: list[str] | None = None,
        provider_ignore: list[str] | None = None,
        quantizations: list[str] | None = None,
        provider_sort: Literal["price", "throughput", "latency"] | None = None,
        max_price: dict[str, float] | None = None,
    ):
        if not (api_key or os.getenv("OPENROUTER_API_KEY")):
            exception_string = "OPENROUTER_API_KEY variable or api_key required."
            raise openai.OpenAIError(exception_string)
        self._openrouter_openai_chat_model = _OpenRouterOpenaiChatModel(
            model,
            api_key=api_key or os.getenv("OPENROUTER_API_KEY"),
            base_url=base_url,
            max_tokens=max_tokens,
            seed=seed,
            temperature=temperature,
            route=route,
            models=models,
            reasoning_effort=reasoning_effort,
            reasoning_exclude=reasoning_exclude,
            require_parameters=require_parameters,
            provider_order=provider_order,
            allow_fallbacks=allow_fallbacks,
            data_collection=data_collection,
            provider_only=provider_only,
            provider_ignore=provider_ignore,
            quantizations=quantizations,
            provider_sort=provider_sort,
            max_price=max_price,
        )

    def _get_extra_body(self) -> dict[str, Any] | None:
        """Get extra body parameters for OpenRouter API."""
        return self._openrouter_openai_chat_model._get_extra_body()

    @property
    def model(self) -> str:
        return self._openrouter_openai_chat_model.model

    @property
    def api_key(self) -> str | None:
        return self._openrouter_openai_chat_model.api_key

    @property
    def base_url(self) -> str | None:
        return self._openrouter_openai_chat_model.base_url

    @property
    def max_tokens(self) -> int | None:
        return self._openrouter_openai_chat_model.max_tokens

    @property
    def seed(self) -> int | None:
        return self._openrouter_openai_chat_model.seed

    @property
    def temperature(self) -> float | None:
        return self._openrouter_openai_chat_model.temperature

    @property
    def route(self) -> Literal["fallback"] | None:
        return self._openrouter_openai_chat_model._route

    @property
    def models(self) -> list[str] | None:
        return self._openrouter_openai_chat_model._models

    @property
    def require_parameters(self) -> bool | None:
        return self._openrouter_openai_chat_model._require_parameters

    @property
    def reasoning(self) -> dict[str, Any] | None:
        return {
            "effort": self._openrouter_openai_chat_model._reasoning_effort,
            "exclude": self._openrouter_openai_chat_model._reasoning_exclude,
        }

    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[OutputT]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> OpenRouterAssistantMessage[OutputT]:
        """Request an LLM message."""
        return self._openrouter_openai_chat_model.complete(
            messages=messages,
            functions=functions,
            output_types=output_types,
            stop=stop,
        )

    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[OutputT]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> OpenRouterAssistantMessage[OutputT]:
        """Async version of `complete`."""
        return await self._openrouter_openai_chat_model.acomplete(
            messages=messages,
            functions=functions,
            output_types=output_types,
            stop=stop,
        )



import json
from collections.abc import AsyncIterator, Callable, Iterable, Iterator, Sequence
from enum import Enum
from functools import singledispatch
from itertools import groupby
from typing import Any, Generic, cast

from typing_extensions import TypeVar

from magentic._parsing import contains_parallel_function_call_type, contains_string_type
from magentic._streamed_response import AsyncStreamedResponse, StreamedResponse
from magentic.chat_model.base import ChatModel, OutputT, aparse_stream, parse_stream
from magentic.chat_model.function_schema import (
    BaseFunctionSchema,
    FunctionCallFunctionSchema,
    function_schema_for_type,
    get_async_function_schemas,
    get_function_schemas,
)
from magentic.chat_model.message import (
    AssistantMessage,
    DocumentBytes,
    ImageBytes,
    Message,
    SystemMessage,
    ToolResultMessage,
    Usage,
    UserMessage,
    _RawMessage,
)
from magentic.chat_model.stream import (
    AsyncOutputStream,
    FunctionCallChunk,
    OutputStream,
    StreamParser,
    StreamState,
)
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
    _create_unique_id,
)
from magentic.streaming import AsyncStreamedStr, StreamedStr
from magentic.vision import UserImageMessage

try:
    import anthropic
    from anthropic.lib.streaming import MessageStreamEvent
    from anthropic.lib.streaming._messages import accumulate_event
    from anthropic.types import (
        DocumentBlockParam,
        ImageBlockParam,
        MessageParam,
        TextBlockParam,
        ToolChoiceParam,
        ToolChoiceToolParam,
        ToolParam,
        ToolUseBlockParam,
    )
except ImportError as error:
    msg = "To use AnthropicChatModel you must install the `anthropic` package using `pip install 'magentic[anthropic]'`."
    raise ImportError(msg) from error


class AnthropicMessageRole(Enum):
    ASSISTANT = "assistant"
    USER = "user"


@singledispatch
def message_to_anthropic_message(message: Message[Any]) -> MessageParam:
    """Convert a Message to an OpenAI message."""
    # TODO: Add instructions for registering new Message type to this error message
    raise NotImplementedError(type(message))


@singledispatch
async def async_message_to_anthropic_message(message: Message[Any]) -> MessageParam:
    """Async version of `message_to_anthropic_message`."""
    return message_to_anthropic_message(message)


@message_to_anthropic_message.register(_RawMessage)
def _(message: _RawMessage[Any]) -> MessageParam:
    # TODO: Validate the message content
    return message.content  # type: ignore[no-any-return]


@message_to_anthropic_message.register(UserMessage)
def _(message: UserMessage[Any]) -> MessageParam:
    if isinstance(message.content, str):
        return {"role": AnthropicMessageRole.USER.value, "content": message.content}
    if isinstance(message.content, Iterable):
        content: list[TextBlockParam | DocumentBlockParam | ImageBlockParam] = []
        for block in message.content:
            if isinstance(block, str):
                content.append({"type": "text", "text": block})
            elif isinstance(block, DocumentBytes):
                content.append(
                    {
                        "type": "document",
                        "source": {
                            "type": "base64",
                            "media_type": block.mime_type,
                            "data": block.as_base64(),
                        },
                    }
                )
            elif isinstance(block, ImageBytes):
                content.append(
                    {
                        "type": "image",
                        "source": {
                            "type": "base64",
                            "media_type": block.mime_type,
                            "data": block.as_base64(),
                        },
                    }
                )
            else:
                msg = f"Invalid content type for UserMessage: {type(block)}"
                raise TypeError(msg)
        return {"role": AnthropicMessageRole.USER.value, "content": content}
    msg = f"Invalid content type for UserMessage: {type(message.content)}"
    raise TypeError(msg)


@message_to_anthropic_message.register(UserImageMessage)
def _(message: UserImageMessage[Any]) -> MessageParam:
    if not isinstance(message.content, bytes):
        msg = f"Invalid content type: {type(message.content)}"
        raise TypeError(msg)

    image_bytes = ImageBytes(message.content)
    return {
        "role": AnthropicMessageRole.USER.value,
        "content": [
            {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": image_bytes.mime_type,
                    "data": image_bytes.as_base64(),
                },
            }
        ],
    }


def _function_call_to_tool_call_block(
    function_call: FunctionCall[Any],
) -> ToolUseBlockParam:
    function_schema = FunctionCallFunctionSchema(function_call.function)
    return {
        "type": "tool_use",
        "id": function_call._unique_id,
        "name": function_schema.name,
        "input": json.loads(function_schema.serialize_args(function_call)),
    }


@message_to_anthropic_message.register(AssistantMessage)
def _(message: AssistantMessage[Any]) -> MessageParam:
    if isinstance(message.content, str):
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": message.content,
        }

    if isinstance(message.content, FunctionCall):
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": [_function_call_to_tool_call_block(message.content)],
        }

    if isinstance(message.content, ParallelFunctionCall):
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": [
                _function_call_to_tool_call_block(function_call)
                for function_call in message.content
            ],
        }

    if isinstance(message.content, StreamedResponse):
        content_blocks: list[TextBlockParam | ToolUseBlockParam] = []
        for item in message.content:
            if isinstance(item, StreamedStr):
                content_blocks.append({"type": "text", "text": item.to_string()})
            elif isinstance(item, FunctionCall):
                content_blocks.append(_function_call_to_tool_call_block(item))
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": content_blocks,
        }

    function_schema = function_schema_for_type(type(message.content))
    return {
        "role": AnthropicMessageRole.ASSISTANT.value,
        "content": [
            {
                "type": "tool_use",
                # Can be random because no result will be inserted back into the chat
                "id": _create_unique_id(),
                "name": function_schema.name,
                "input": json.loads(function_schema.serialize_args(message.content)),
            }
        ],
    }


@async_message_to_anthropic_message.register(AssistantMessage)
async def _(message: AssistantMessage[Any]) -> MessageParam:
    if isinstance(message.content, AsyncParallelFunctionCall):
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": [
                _function_call_to_tool_call_block(function_call)
                async for function_call in message.content
            ],
        }

    if isinstance(message.content, AsyncStreamedResponse):
        content_blocks: list[TextBlockParam | ToolUseBlockParam] = []
        async for item in message.content:
            if isinstance(item, AsyncStreamedStr):
                content_blocks.append({"type": "text", "text": await item.to_string()})
            elif isinstance(item, FunctionCall):
                content_blocks.append(_function_call_to_tool_call_block(item))
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": content_blocks,
        }
    return message_to_anthropic_message(message)


@message_to_anthropic_message.register(ToolResultMessage)
def _(message: ToolResultMessage[Any]) -> MessageParam:
    if isinstance(message.content, str):
        content = message.content
    else:
        function_schema = function_schema_for_type(type(message.content))
        content = json.loads(function_schema.serialize_args(message.content))
    return {
        "role": AnthropicMessageRole.USER.value,
        "content": [
            {
                "type": "tool_result",
                "tool_use_id": message.tool_call_id,
                "content": content,
            }
        ],
    }


# TODO: Move this to the magentic level by allowing `UserMessage` have a list of content
def _combine_messages(messages: Iterable[MessageParam]) -> list[MessageParam]:
    """Combine messages with the same role, to get alternating roles.

    Alternating roles is a requirement of the Anthropic API.
    """
    combined_messages: list[MessageParam] = []
    for message_group in groupby(messages, lambda x: x["role"]):
        role, messages = message_group
        content = []
        for message in messages:
            if isinstance(message["content"], list):
                content.extend(message["content"])
            elif isinstance(message["content"], str):
                content.append({"type": "text", "text": message["content"]})
            else:
                content.append(message["content"])
        combined_messages.append({"role": role, "content": content})
    return combined_messages


T = TypeVar("T")
BaseFunctionSchemaT = TypeVar("BaseFunctionSchemaT", bound=BaseFunctionSchema[Any])


class BaseFunctionToolSchema(Generic[BaseFunctionSchemaT]):
    def __init__(self, function_schema: BaseFunctionSchemaT):
        self._function_schema = function_schema

    def to_dict(self) -> ToolParam:
        return {
            "name": self._function_schema.name,
            "description": self._function_schema.description or "",
            "input_schema": self._function_schema.parameters,
        }

    def as_tool_choice(self, *, disable_parallel_tool_use: bool) -> ToolChoiceToolParam:
        return {"type": "tool", "name": self._function_schema.name}


class AnthropicStreamParser(StreamParser[MessageStreamEvent]):
    def is_content(self, item: MessageStreamEvent) -> bool:
        return item.type == "content_block_delta"

    def get_content(self, item: MessageStreamEvent) -> str | None:
        if item.type == "text":
            return item.text
        return None

    def is_tool_call(self, item: MessageStreamEvent) -> bool:
        return (
            item.type == "content_block_start" and item.content_block.type == "tool_use"
        )

    def iter_tool_calls(self, item: MessageStreamEvent) -> Iterable[FunctionCallChunk]:
        if item.type == "content_block_start" and item.content_block.type == "tool_use":
            return [
                FunctionCallChunk(
                    id=item.content_block.id, name=item.content_block.name, args=None
                )
            ]
        if item.type == "input_json":
            return [FunctionCallChunk(id=None, name=None, args=item.partial_json)]
        return []


class AnthropicStreamState(StreamState[MessageStreamEvent]):
    def __init__(self) -> None:
        self._current_message_snapshot: anthropic.types.Message | None = (
            None  # TODO: type
        )
        self.usage_ref: list[Usage] = []

    def update(self, item: MessageStreamEvent) -> None:
        self._current_message_snapshot = accumulate_event(
            # Unrecognized event types are ignored
            event=item,  # type: ignore[arg-type]
            current_snapshot=self._current_message_snapshot,
        )
        if item.type == "message_stop":
            assert not self.usage_ref
            self.usage_ref.append(
                Usage(
                    input_tokens=item.message.usage.input_tokens,
                    output_tokens=item.message.usage.output_tokens,
                )
            )

    @property
    def current_message_snapshot(self) -> Message[Any]:
        assert self._current_message_snapshot is not None
        # TODO: Possible to return AssistantMessage here?
        return _RawMessage(self._current_message_snapshot.model_dump())


def _extract_system_message(
    messages: Iterable[Message[Any]],
) -> tuple[str | anthropic.NotGiven, list[Message[Any]]]:
    system_messages = [m for m in messages if isinstance(m, SystemMessage)]
    if len(system_messages) > 1:
        msg = "Only one system message is allowed per request."
        raise ValueError(msg)
    return (
        system_messages[0].content if system_messages else anthropic.NOT_GIVEN,
        [m for m in messages if not isinstance(m, SystemMessage)],
    )


def _if_given(value: T | None) -> T | anthropic.NotGiven:
    return value if value is not None else anthropic.NOT_GIVEN


class AnthropicChatModel(ChatModel):
    """An LLM chat model that uses the `anthropic` python package."""

    def __init__(
        self,
        model: str,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
        max_tokens: int = 1024,
        temperature: float | None = None,
    ):
        self._model = model
        self._api_key = api_key
        self._base_url = base_url
        self._max_tokens = max_tokens
        self._temperature = temperature

        self._client = anthropic.Anthropic(api_key=api_key, base_url=base_url)
        self._async_client = anthropic.AsyncAnthropic(
            api_key=api_key, base_url=base_url
        )

    @property
    def model(self) -> str:
        return self._model

    @property
    def api_key(self) -> str | None:
        return self._api_key

    @property
    def base_url(self) -> str | None:
        return self._base_url

    @property
    def max_tokens(self) -> int:
        return self._max_tokens

    @property
    def temperature(self) -> float | None:
        return self._temperature

    @staticmethod
    def _get_tool_choice(
        *,
        tool_schemas: Sequence[BaseFunctionToolSchema[Any]],
        output_types: Iterable[type],
    ) -> ToolChoiceParam | anthropic.NotGiven:
        """Create the tool choice argument."""
        if contains_string_type(output_types):
            return anthropic.NOT_GIVEN
        disable_parallel_tool_use = not contains_parallel_function_call_type(
            output_types
        )
        if len(tool_schemas) == 1:
            return tool_schemas[0].as_tool_choice(
                disable_parallel_tool_use=disable_parallel_tool_use
            )
        return {"type": "any", "disable_parallel_tool_use": disable_parallel_tool_use}

    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[OutputT]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[OutputT]:
        """Request an LLM message."""
        if output_types is None:
            output_types = [] if functions else cast(list[type[OutputT]], [str])

        function_schemas = get_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        system, messages = _extract_system_message(messages)

        response: Iterator[MessageStreamEvent] = self._client.messages.stream(
            model=self.model,
            messages=_combine_messages(
                [message_to_anthropic_message(m) for m in messages]
            ),
            max_tokens=self.max_tokens,
            stop_sequences=_if_given(stop),
            system=system,
            temperature=_if_given(self.temperature),
            tools=[schema.to_dict() for schema in tool_schemas] or anthropic.NOT_GIVEN,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),
        ).__enter__()
        stream = OutputStream(
            response,
            function_schemas=function_schemas,
            parser=AnthropicStreamParser(),
            state=AnthropicStreamState(),
        )
        return AssistantMessage._with_usage(
            parse_stream(stream, output_types), usage_ref=stream.usage_ref
        )

    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[OutputT]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[OutputT]:
        """Async version of `complete`."""
        if output_types is None:
            output_types = [] if functions else cast(list[type[OutputT]], [str])

        function_schemas = get_async_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        system, messages = _extract_system_message(messages)

        response: AsyncIterator[
            MessageStreamEvent
        ] = await self._async_client.messages.stream(
            model=self.model,
            messages=_combine_messages(
                [await async_message_to_anthropic_message(m) for m in messages]
            ),
            max_tokens=self.max_tokens,
            stop_sequences=_if_given(stop),
            system=system,
            temperature=_if_given(self.temperature),
            tools=[schema.to_dict() for schema in tool_schemas] or anthropic.NOT_GIVEN,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),
        ).__aenter__()
        stream = AsyncOutputStream(
            response,
            function_schemas=function_schemas,
            parser=AnthropicStreamParser(),
            state=AnthropicStreamState(),
        )
        return AssistantMessage._with_usage(
            await aparse_stream(stream, output_types), usage_ref=stream.usage_ref
        )



import os
from typing import Annotated

import openai
import pytest
from pydantic import AfterValidator, BaseModel

from magentic._pydantic import ConfigDict, with_config
from magentic._streamed_response import AsyncStreamedResponse, StreamedResponse
from magentic.chat_model.base import ToolSchemaParseError
from magentic.chat_model.message import ImageBytes, Message, Usage, UserMessage
from magentic.chat_model.openrouter_chat_model import OpenRouterChatModel
from magentic.function_call import FunctionCall
from magentic.streaming import AsyncStreamedStr, StreamedStr


@pytest.mark.openrouter
def test_openrouter_chat_model_api_key(monkeypatch):
    openrouter_api_key = os.environ["OPENROUTER_API_KEY"]
    monkeypatch.delenv("OPENROUTER_API_KEY")

    with pytest.raises(openai.OpenAIError):
        chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")

    chat_model = OpenRouterChatModel(
        "deepseek/deepseek-chat-v3-0324", api_key=openrouter_api_key
    )
    message = chat_model.complete(messages=[UserMessage("Say hello!")])
    assert isinstance(message.content, str)


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_base_url():
    chat_model = OpenRouterChatModel(
        "deepseek/deepseek-chat-v3-0324", base_url="https://openrouter.ai/api/v1"
    )
    message = chat_model.complete(messages=[UserMessage("Say hello!")])
    assert isinstance(message.content, str)


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_seed():
    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324", seed=42)
    message1 = chat_model.complete(messages=[UserMessage("Say hello!")])
    message2 = chat_model.complete(messages=[UserMessage("Say hello!")])
    # Check that both responses are greetings
    assert any(
        greeting in message1.content.lower() for greeting in ["hello", "hi", "hey"]
    )
    assert any(
        greeting in message2.content.lower() for greeting in ["hello", "hi", "hey"]
    )


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_streamed_response():
    def get_weather(location: str) -> None:
        """Get the weather for a location."""

    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")
    message = chat_model.complete(
        messages=[UserMessage("Tell me your favorite city. Then get its weather.")],
        functions=[get_weather],
        output_types=[StreamedResponse],
    )
    assert isinstance(message.content, StreamedResponse)
    response_items = list(message.content)
    assert len(response_items) == 2
    streamed_str, function_call = response_items
    assert isinstance(streamed_str, StreamedStr)
    assert len(streamed_str.to_string()) > 1  # Check StreamedStr was cached
    assert isinstance(function_call, FunctionCall)
    assert function_call() is None  # Check FunctionCall is successfully called


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_pydantic_model_openai_strict():
    class CapitalCity(BaseModel):
        model_config = ConfigDict(openai_strict=True)
        capital: str
        country: str

    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")
    message = chat_model.complete(
        messages=[UserMessage("What is the capital of Ireland?")],
        output_types=[CapitalCity],
    )
    assert isinstance(message.content, CapitalCity)


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_function_call_openai_strict():
    @with_config(ConfigDict(openai_strict=True))
    def plus(a: int, b: int) -> int:
        """Sum two numbers."""
        return a + b

    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")
    message = chat_model.complete(
        messages=[UserMessage("Use the tool to sum 1 and 2")],
        functions=[plus],
        output_types=[FunctionCall[int]],
    )
    assert isinstance(message.content, FunctionCall)


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_usage():
    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")
    message = chat_model.complete(
        messages=[UserMessage("Say hello!")], output_types=[StreamedStr]
    )
    str(message.content)  # Finish the stream
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_usage_structured_output():
    chat_model = OpenRouterChatModel(
        "deepseek/deepseek-chat-v3-0324", require_parameters=True
    )
    message = chat_model.complete(
        messages=[UserMessage("Count to 5. Tool call please.")],
        output_types=[list[int]],
    )
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_no_structured_output_error():
    chat_model = OpenRouterChatModel(
        "deepseek/deepseek-chat-v3-0324",
        require_parameters=True,  # This will force tool calls
    )
    message: Message[int | bool] = chat_model.complete(
        messages=[UserMessage("What is 2+2? Integer please.")],
        output_types=[int, bool],
    )
    assert isinstance(message.content, int | bool)


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_raises_tool_schema_parse_error():
    def raise_error(v):
        raise ValueError(v)

    class Test(BaseModel):
        value: Annotated[int, AfterValidator(raise_error)]

    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")
    with pytest.raises(ToolSchemaParseError):
        chat_model.complete(
            messages=[UserMessage("Return a test value of 42.")],
            output_types=[Test],
        )


@pytest.mark.openrouter
def test_openrouter_chat_model_complete_image_bytes(image_bytes_jpg):
    chat_model = OpenRouterChatModel(
        "opengvlab/internvl3-14b:free", require_parameters=True
    )
    message = chat_model.complete(
        messages=[
            UserMessage(
                ("Describe this image in one word.", ImageBytes(image_bytes_jpg))
            )
        ]
    )
    assert isinstance(message.content, str)


@pytest.mark.openrouter
async def test_openrouter_chat_model_acomplete_async_streamed_response():
    def get_weather(location: str) -> None:
        """Get the weather for a location."""

    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")
    message = await chat_model.acomplete(
        messages=[UserMessage("Tell me your favorite city. Then get its weather.")],
        functions=[get_weather],
        output_types=[AsyncStreamedResponse],
    )
    assert isinstance(message.content, AsyncStreamedResponse)
    response_items = [x async for x in message.content]
    assert len(response_items) == 2
    streamed_str, function_call = response_items
    assert isinstance(streamed_str, AsyncStreamedStr)
    assert len(await streamed_str.to_string()) > 1  # Check AsyncStreamedStr was cached
    assert isinstance(function_call, FunctionCall)
    assert function_call() is None  # Check FunctionCall is successfully called


@pytest.mark.openrouter
async def test_openrouter_chat_model_acomplete_usage():
    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")
    message = await chat_model.acomplete(
        messages=[UserMessage("Say hello!")], output_types=[AsyncStreamedStr]
    )
    await message.content.to_string()  # Finish the stream
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.openrouter
async def test_openrouter_chat_model_acomplete_usage_structured_output():
    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")
    message = await chat_model.acomplete(
        messages=[UserMessage("Count to 5")], output_types=[list[int]]
    )
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.openrouter
async def test_openrouter_chat_model_acomplete_raises_tool_schema_parse_error():
    def raise_error(v):
        raise ValueError(v)

    class Test(BaseModel):
        value: Annotated[int, AfterValidator(raise_error)]

    chat_model = OpenRouterChatModel("deepseek/deepseek-chat-v3-0324")
    with pytest.raises(ToolSchemaParseError):
        await chat_model.acomplete(
            messages=[UserMessage("Return a test value of 42.")],
            output_types=[Test],
        )


@pytest.mark.openrouter
def test_openrouter_chat_model_extra_body():
    chat_model = OpenRouterChatModel(
        "deepseek/deepseek-chat-v3-0324",
        route="fallback",
        models=["anthropic/claude-3-opus"],
        require_parameters=True,
        reasoning_effort="low",
        reasoning_exclude=False,
        provider_only=["Novita"],
    )
    assert chat_model._get_extra_body() == {
        "route": "fallback",
        "models": ["anthropic/claude-3-opus"],
        "provider": {"require_parameters": True, "only": ["Novita"]},
        "reasoning": {"effort": "low", "exclude": False},
    }


@pytest.mark.openrouter
def test_openrouter_chat_model_reasoning_tokens():
    chat_model = OpenRouterChatModel(
        "deepseek/deepseek-r1",
        require_parameters=True,
        reasoning_effort="low",
        reasoning_exclude=False,
    )
    message = chat_model.complete(
        messages=[UserMessage("Tell me a joke. Make it catchy.")],
        output_types=[str],
    )

    # Verify response content
    assert isinstance(message.content, str)
    assert len(message.content) > 0

    # Verify reasoning tokens are present in the response
    assert hasattr(message, "reasoning")
    assert message.reasoning is not None
    assert isinstance(message.reasoning, str)
    assert len(message.reasoning) > 0

    # Verify usage stats are present
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0



import os

import pytest

from magentic.chat_model.message import ImageBytes, UserMessage
from magentic.chat_model.openai_chat_model import OpenaiChatModel


@pytest.fixture
def chat_model():
    return OpenaiChatModel(
        "grok-2", base_url="https://api.x.ai/v1", api_key=os.environ["XAI_API_KEY"]
    )


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True.", [bool], bool),
        ("Return [1, 2, 3, 4, 5]", [list[int]], list),
        ("Return a list of fruit", [list[str]], list),
    ],
)
@pytest.mark.openai_xai
def test_openai_chat_model_xai_complete(
    chat_model, prompt, output_types, expected_output_type
):
    message = chat_model.complete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True.", [bool], bool),
        ("Return [1, 2, 3, 4, 5]", [list[int]], list),
        ("Return a list of fruit", [list[str]], list),
    ],
)
@pytest.mark.openai_xai
async def test_openai_chat_model_xai_acomplete(
    chat_model, prompt, output_types, expected_output_type
):
    message = await chat_model.acomplete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.mark.openai_xai
def test_openai_chat_model_xai_complete_image_bytes(image_bytes_jpg):
    chat_model = OpenaiChatModel(
        "grok-2-vision",
        base_url="https://api.x.ai/v1",
        api_key=os.environ["XAI_API_KEY"],
    )
    message = chat_model.complete(
        messages=[
            UserMessage(
                ("Describe this image in one word.", ImageBytes(image_bytes_jpg))
            )
        ]
    )
    assert isinstance(message.content, str)



import inspect
import typing
from abc import ABC, abstractmethod
from collections.abc import AsyncIterable, Callable, Iterable
from functools import singledispatch
from typing import Any, Generic, TypeVar, cast, get_args, get_origin

from openai.types.shared_params import FunctionDefinition
from pydantic import BaseModel, TypeAdapter, create_model

from magentic._pydantic import ConfigDict, get_pydantic_config, json_schema
from magentic._streamed_response import AsyncStreamedResponse, StreamedResponse
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
)
from magentic.streaming import (
    AsyncStreamedStr,
    StreamedStr,
    aiter_streamed_json_array,
    iter_streamed_json_array,
)
from magentic.typing import is_origin_abstract, is_origin_subclass, name_type

T = TypeVar("T")


class BaseFunctionSchema(ABC, Generic[T]):
    """Converts a Python object to the JSON Schema that represents it as a function for the LLM."""

    # Allow any arguments to avoid error passing type to subclasses without __init__
    def __init__(self, *args: Any, **kwargs: Any): ...

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the function.

        Must be a-z, A-Z, 0-9, or contain underscores and dashes, with a maximum length of 64.
        """
        ...

    @property
    def description(self) -> str | None:
        """A description of what the function does."""
        return None

    @property
    @abstractmethod
    def parameters(self) -> dict[str, Any]:
        """The parameters the functions accepts as a JSON Schema object."""
        ...

    @property
    def strict(self) -> bool | None:
        """Whether to enable strict schema adherence when generating the function call."""
        return None

    def dict(self) -> FunctionDefinition:
        schema: FunctionDefinition = {"name": self.name, "parameters": self.parameters}
        if self.description:
            schema["description"] = self.description
        if self.strict is not None:
            schema["strict"] = self.strict
        return schema


BaseFunctionSchemaT = TypeVar("BaseFunctionSchemaT", bound=BaseFunctionSchema[Any])


def select_function_schema(
    function_schemas: Iterable[BaseFunctionSchemaT], name: str
) -> BaseFunctionSchemaT | None:
    """Select the function schema with the given name."""
    for schema in function_schemas:
        if schema.name == name:
            return schema
    return None


class AsyncFunctionSchema(BaseFunctionSchema[T], Generic[T]):
    @abstractmethod
    async def aparse_args(self, chunks: AsyncIterable[str]) -> T:
        """Parse an async iterable of string chunks into the function arguments."""
        ...

    @abstractmethod
    async def aserialize_args(self, value: T) -> str:
        """Serialize the function arguments into a JSON string."""
        ...


class FunctionSchema(AsyncFunctionSchema[T], Generic[T]):
    @abstractmethod
    def parse_args(self, chunks: Iterable[str]) -> T:
        """Parse an iterable of string chunks into the function arguments."""
        ...

    @abstractmethod
    def serialize_args(self, value: T) -> str:
        """Serialize the function arguments into a JSON string."""
        ...

    async def aparse_args(self, chunks: AsyncIterable[str]) -> T:
        """Parse an async iterable of string chunks into the function arguments."""
        return self.parse_args([chunk async for chunk in chunks])

    async def aserialize_args(self, value: T) -> str:
        """Serialize the function arguments into a JSON string."""
        return self.serialize_args(value)


# Use the singledispatch registry to map classes to FunctionSchemas
# because this handles subclass resolution for us.
@singledispatch
def _async_function_schema_registry(type_: type[T]) -> AsyncFunctionSchema[T]:
    msg = f"No FunctionSchema registered for type {type_}"
    raise TypeError(msg)


def async_function_schema_for_type(type_: type[T]) -> AsyncFunctionSchema[T]:
    """Create a FunctionSchema for the given type."""
    function_schema_cls = _async_function_schema_registry.dispatch(
        get_origin(type_) or type_
    )
    return function_schema_cls(type_)


@singledispatch
def _function_schema_registry(type_: type[T]) -> FunctionSchema[T]:
    msg = f"No FunctionSchema registered for type {type_}"
    raise TypeError(msg)


def function_schema_for_type(type_: type[T]) -> FunctionSchema[T]:
    """Create a FunctionSchema for the given type."""
    function_schema_cls = _function_schema_registry.dispatch(get_origin(type_) or type_)
    return function_schema_cls(type_)


TypeFunctionSchemaT = TypeVar(
    "TypeFunctionSchemaT", bound=type[BaseFunctionSchema[Any]]
)


def register_function_schema(
    type_: type[Any],
) -> Callable[[TypeFunctionSchemaT], TypeFunctionSchemaT]:
    """Register a new FunctionSchema for the given type."""

    def _register(cls: TypeFunctionSchemaT) -> TypeFunctionSchemaT:
        if cls.__abstractmethods__:
            msg = f"FunctionSchema {cls} has not implemented abstract methods: {cls.__abstractmethods__}"
            raise TypeError(msg)

        if issubclass(cls, AsyncFunctionSchema):
            _async_function_schema_registry.register(type_, cls)
        if issubclass(cls, FunctionSchema):
            _function_schema_registry.register(type_, cls)
        return cls

    return _register


@register_function_schema(object)
class AnyFunctionSchema(FunctionSchema[T], Generic[T]):
    """The most generic FunctionSchema that should work for most types supported by pydantic."""

    def __init__(self, output_type: type[T]):
        self._output_type = output_type
        self._model = create_model(
            "Output",
            __config__=get_pydantic_config(output_type),
            value=(output_type, ...),
        )

    @property
    def name(self) -> str:
        return f"return_{name_type(self._output_type)}"

    @property
    def parameters(self) -> dict[str, Any]:
        return json_schema(self._model)

    @property
    def strict(self) -> bool | None:
        return cast(ConfigDict, self._model.model_config).get("openai_strict")

    def parse_args(self, chunks: Iterable[str]) -> T:
        args_json = "".join(chunks)
        return cast(T, self._model.model_validate_json(args_json).value)  # type: ignore[attr-defined]

    def serialize_args(self, value: T) -> str:
        return self._model.model_construct(value=value).model_dump_json()


IterableT = TypeVar("IterableT", bound=Iterable[Any])


@register_function_schema(Iterable)
class IterableFunctionSchema(FunctionSchema[IterableT], Generic[IterableT]):
    """FunctionSchema for types that are iterable. Can parse LLM output as a stream."""

    def __init__(self, output_type: type[IterableT]):
        self._output_type = output_type
        self._item_type_adapter: TypeAdapter[Any] = TypeAdapter(
            args[0] if (args := get_args(output_type)) else Any
        )
        self._model = create_model(
            "Output",
            __config__=get_pydantic_config(output_type),
            value=(output_type, ...),
        )

    @property
    def name(self) -> str:
        return f"return_{name_type(self._output_type)}"

    @property
    def parameters(self) -> dict[str, Any]:
        return json_schema(self._model)

    @property
    def strict(self) -> bool | None:
        return cast(ConfigDict, self._model.model_config).get("openai_strict")

    def parse_args(self, chunks: Iterable[str]) -> IterableT:
        iter_items = (
            self._item_type_adapter.validate_json(item)
            for item in iter_streamed_json_array(chunks)
        )
        return cast(IterableT, self._model.model_validate({"value": iter_items}).value)  # type: ignore[attr-defined]

    def serialize_args(self, value: IterableT) -> str:
        return self._model.model_construct(value=value).model_dump_json()


AsyncIterableT = TypeVar("AsyncIterableT", bound=AsyncIterable[Any])


@register_function_schema(AsyncIterable)
class AsyncIterableFunctionSchema(
    AsyncFunctionSchema[AsyncIterableT], Generic[AsyncIterableT]
):
    """FunctionSchema for types that are async iterable. Can parse LLM output as a stream."""

    def __init__(self, output_type: type[AsyncIterableT]):
        self._output_type = output_type
        item_type: type | Any = args[0] if (args := get_args(output_type)) else Any
        self._item_type_adapter = TypeAdapter(item_type)
        self._model = create_model(
            "Output",
            __config__=get_pydantic_config(output_type),
            # Convert to list so pydantic can handle for schema generation
            value=(list[item_type], ...),  # type: ignore[valid-type]
        )

    @property
    def name(self) -> str:
        return f"return_{name_type(self._output_type)}"

    @property
    def parameters(self) -> dict[str, Any]:
        return json_schema(self._model)

    @property
    def strict(self) -> bool | None:
        return cast(ConfigDict, self._model.model_config).get("openai_strict")

    async def aparse_args(self, chunks: AsyncIterable[str]) -> AsyncIterableT:
        aiter_items = (
            self._item_type_adapter.validate_json(item)
            async for item in aiter_streamed_json_array(chunks)
        )
        if (get_origin(self._output_type) or self._output_type) in (
            typing.AsyncIterable,
            typing.AsyncIterator,
        ) or is_origin_abstract(self._output_type):
            return cast(AsyncIterableT, aiter_items)

        raise NotImplementedError

    async def aserialize_args(self, value: AsyncIterableT) -> str:
        return self._model.model_construct(
            value=[chunk async for chunk in value]
        ).model_dump_json()


@register_function_schema(dict)
class DictFunctionSchema(FunctionSchema[T], Generic[T]):
    """FunctionSchema for dict."""

    def __init__(self, output_type: type[T]):
        self._output_type = output_type
        self._type_adapter = TypeAdapter(
            output_type, config=get_pydantic_config(output_type)
        )

    @property
    def name(self) -> str:
        return f"return_{name_type(self._output_type)}"

    @property
    def parameters(self) -> dict[str, Any]:
        model_schema = self._type_adapter.json_schema().copy()
        model_schema["properties"] = model_schema.get("properties", {})
        return model_schema

    def parse_args(self, chunks: Iterable[str]) -> T:
        args_json = "".join(chunks)
        return self._type_adapter.validate_json(args_json)

    def serialize_args(self, value: T) -> str:
        return self._type_adapter.dump_json(value).decode()


BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


@register_function_schema(BaseModel)
class BaseModelFunctionSchema(FunctionSchema[BaseModelT], Generic[BaseModelT]):
    """FunctionSchema for pydantic BaseModel."""

    def __init__(self, model: type[BaseModelT]):
        self._model = model

    @property
    def name(self) -> str:
        return f"return_{name_type(self._model)}"

    @property
    def parameters(self) -> dict[str, Any]:
        return json_schema(self._model)

    @property
    def strict(self) -> bool | None:
        return cast(ConfigDict, self._model.model_config).get("openai_strict")

    def parse_args(self, chunks: Iterable[str]) -> BaseModelT:
        args_json = "".join(chunks)
        return self._model.model_validate_json(args_json)

    def serialize_args(self, value: BaseModelT) -> str:
        return value.model_dump_json()


def create_model_from_function(func: Callable[..., Any]) -> type[BaseModel]:
    """Create a Pydantic model from a function signature."""
    # https://github.com/pydantic/pydantic/issues/3585#issuecomment-1002745763
    fields: dict[str, Any] = {}
    for param in inspect.signature(func).parameters.values():
        # *args
        if param.kind is inspect.Parameter.VAR_POSITIONAL:
            fields[param.name] = (
                (
                    list[param.annotation]  # type: ignore[name-defined]
                    if param.annotation != inspect._empty
                    else list[Any]
                ),
                param.default if param.default != inspect._empty else [],
            )
            continue

        # **kwargs
        if param.kind is inspect.Parameter.VAR_KEYWORD:
            fields[param.name] = (
                dict[str, param.annotation]  # type: ignore[name-defined]
                if param.annotation != inspect._empty
                else dict[str, Any],
                param.default if param.default != inspect._empty else {},
            )
            continue

        fields[param.name] = (
            (param.annotation if param.annotation != inspect._empty else Any),
            (param.default if param.default != inspect._empty else ...),
        )
    return create_model("FuncModel", __config__=get_pydantic_config(func), **fields)


class FunctionCallFunctionSchema(FunctionSchema[FunctionCall[T]], Generic[T]):
    """FunctionSchema for FunctionCall."""

    def __init__(self, func: Callable[..., T]):
        self._func = func
        self._model = create_model_from_function(func)

    @property
    def name(self) -> str:
        return self._func.__name__

    @property
    def description(self) -> str | None:
        return inspect.getdoc(self._func)

    @property
    def parameters(self) -> dict[str, Any]:
        return json_schema(self._model)

    @property
    def strict(self) -> bool | None:
        return cast(ConfigDict, self._model.model_config).get("openai_strict")

    def parse_args(self, chunks: Iterable[str]) -> FunctionCall[T]:
        # Anthropic message stream returns empty string for function call with no arguments
        args_json = "".join(chunks) or "{}"
        model = self._model.model_validate_json(args_json)
        supplied_params = [
            param
            for param in inspect.signature(self._func).parameters.values()
            if param.name in model.model_fields_set
        ]
        # Prefer keyword args when possible (no *args present)
        # so that default values are used for keyword args that are not provided
        has_var_positional = any(
            param.kind == param.VAR_POSITIONAL for param in supplied_params
        )

        args_positional_only = [
            getattr(model, param.name)
            for param in supplied_params
            if param.kind == param.POSITIONAL_ONLY
        ]
        args_positional_or_keyword_using_position = (
            [
                getattr(model, param.name)
                for param in supplied_params
                if param.kind == param.POSITIONAL_OR_KEYWORD
            ]
            if has_var_positional
            else []
        )
        args_var_positional = [
            arg
            for param in supplied_params
            if param.kind == param.VAR_POSITIONAL
            for arg in getattr(model, param.name)
        ]
        args_positional_or_keyword_using_keyword = (
            {
                param.name: getattr(model, param.name)
                for param in supplied_params
                if param.kind == param.POSITIONAL_OR_KEYWORD
            }
            if not has_var_positional
            else {}
        )
        args_keyword_only = {
            param.name: getattr(model, param.name)
            for param in supplied_params
            if param.kind == param.KEYWORD_ONLY
        }
        args_var_keyword = {
            name: value
            for param in supplied_params
            if param.kind == param.VAR_KEYWORD
            for name, value in getattr(model, param.name).items()
        }
        return FunctionCall(
            self._func,
            *args_positional_only,
            *args_positional_or_keyword_using_position,
            *args_var_positional,
            **args_positional_or_keyword_using_keyword,
            **args_keyword_only,
            **args_var_keyword,
        )

    def serialize_args(self, value: FunctionCall[T]) -> str:
        return self._model.model_construct(**value.arguments).model_dump_json(
            exclude_unset=True
        )


R = TypeVar("R")

_NON_FUNCTION_CALL_TYPES = (
    str,
    StreamedStr,
    AsyncStreamedStr,
    FunctionCall,
    ParallelFunctionCall,
    AsyncParallelFunctionCall,
    StreamedResponse,
    AsyncStreamedResponse,
)


def get_function_schemas(
    functions: Iterable[Callable[..., R]] | None,
    output_types: Iterable[type[T]],
) -> Iterable[FunctionSchema[FunctionCall[R] | T]]:
    return [
        *(FunctionCallFunctionSchema(f) for f in functions or []),  # type: ignore[list-item]
        *(
            function_schema_for_type(type_)
            for type_ in output_types
            if not is_origin_subclass(type_, _NON_FUNCTION_CALL_TYPES)  # type: ignore[list-item]
        ),
    ]


def get_async_function_schemas(
    functions: Iterable[Callable[..., R]] | None,
    output_types: Iterable[type[T]],
) -> Iterable[FunctionSchema[FunctionCall[R] | T]]:
    return [
        *(FunctionCallFunctionSchema(f) for f in functions or []),  # type: ignore[list-item]
        *(
            async_function_schema_for_type(type_)
            for type_ in output_types
            if not is_origin_subclass(type_, _NON_FUNCTION_CALL_TYPES)  # type: ignore[list-item]
        ),
    ]
